// https://github.com/plepers/nanogl-pbr/blob/master/bin/nanoglibl

// Native
const fs = require('fs');

// Arguments
const {
	input,
	output,
} = require('./argsHandler');

// Child process
const createProcess = require('./createProcess');

// Constants
const {
	CMFT_DEFAULT_FLAGS,
	CMFT_TOOL_PATH,
	HDR2PNG_DEFAULT_FLAGS,
	HDR2PNG_TOOL_PATH,
	MONTAGE_DEFAULT_FLAGS,
	MONTAGE_TOOL_PATH,
} = require('./constants');

function convertSHToBin() {
	return new Promise((resolve, reject) => {
		fs.readFile(`${output}/sh.c`, 'utf8', (error, data) => {
			if (error) {
				console.error(error);
			}

			const floatArray = data
				// Match everything between first { and last }
				.match(/{(?:\n|.)*}/)[0]

				// Strip all comments
				.replace(/\/\*[\s\S]*?\*\/|([^:]|^)\/\/.*$/g, '')

				// Match all floats
				.match(/[+-]?\d+(\.\d+)?/g);

			// Only grab the band 0, band 1 and band 2
			const result = new Float32Array(floatArray.slice(0, 27));
			const buffer = Buffer.from(new Uint8Array(result.buffer));

			fs.writeFile(`${output}/sh.bin`, buffer, (error) => {
				if (error) {
					reject(console.error(error));
				}

				resolve(console.log(`Written SH to ${output}/sh.bin`));
			});
		});
	});
}

function getGloss(angle) {
	return Math.log(Math.log(0.00001) / Math.log(Math.cos(angle / 180.0 * Math.PI))) / Math.log(2);
}

function generateHighQualityLevel() {
	return new Promise((resolve, reject) => {
		const CMFT_CUSTOM_FLAGS = [
			'--input', `${input}`,
			'--srcFaceSize', '256',
			'--dstFaceSize', '256',
			'--outputNum', '1',
			'--output0', `${output}/env_hi`,
			'--output0params', 'hdr,rgbe,octant',
		];

		const HDR2PNG_CUSTOM_FLAGS = [
			'-v',
			'-i', `${output}/env_hi.hdr`,
			'-o', `${output}/env_hi.png`,
		];

		const CMFT_FLAGS = [...CMFT_DEFAULT_FLAGS, ...CMFT_CUSTOM_FLAGS];
		const HDR2PNG_FLAGS = [...HDR2PNG_DEFAULT_FLAGS, ...HDR2PNG_CUSTOM_FLAGS];

		// CMFT
		createProcess(CMFT_TOOL_PATH, CMFT_FLAGS)
			.then(() => {
				// HDR2PNG
				createProcess(HDR2PNG_TOOL_PATH, HDR2PNG_FLAGS)
					.then(() => {
						resolve();
					})
					.catch((error) => {
						reject(error);
					});
			});
	});
}

function generateBaseLevel() {
	return new Promise((resolve, reject) => {
		const CMFT_CUSTOM_FLAGS = [
			'--input', `${input}`,
			'--srcFaceSize', '128',
			'--dstFaceSize', '128',
			'--outputNum', '1',
			'--output0', `${output}/level0`,
			'--output0params', 'hdr,rgbe,octant',
		];

		const HDR2PNG_CUSTOM_FLAGS = [
			'-v',
			'-i', `${output}/level0.hdr`,
			'-o', `${output}/level0.png`,
		];

		const CMFT_FLAGS = [...CMFT_DEFAULT_FLAGS, ...CMFT_CUSTOM_FLAGS];
		const HDR2PNG_FLAGS = [...HDR2PNG_DEFAULT_FLAGS, ...HDR2PNG_CUSTOM_FLAGS];

		// CMFT
		createProcess(CMFT_TOOL_PATH, CMFT_FLAGS)
			.then(() => {
				// HDR2PNG
				createProcess(HDR2PNG_TOOL_PATH, HDR2PNG_FLAGS)
					.then(() => {
						resolve();
					})
					.catch((error) => {
						reject(error);
					});
			});
	});
}

function generateGlossLevel(level, gloss = false) {
	return new Promise((resolve, reject) => {
		const glossBias = gloss || getGloss(10 + 80 / 7 * level);

		console.log(glossBias);

		const CMFT_CUSTOM_FLAGS = [
			'--input', `${input}`,
			'--filter', 'radiance',
			'--lightingModel', 'blinn',
			'--glossScale', '0',
			'--glossBias', `${glossBias}`,
			'--srcFaceSize', '128',
			'--dstFaceSize', '128',
			'--outputNum', '1',
			'--output0', `${output}/level${level}`,
			'--output0params', 'hdr,rgbe,octant',
		];

		const HDR2PNG_CUSTOM_FLAGS = [
			'-v',
			'-i', `${output}/level${level}.hdr`,
			'-o', `${output}/level${level}.png`,
		];

		const CMFT_FLAGS = [...CMFT_DEFAULT_FLAGS, ...CMFT_CUSTOM_FLAGS];
		const HDR2PNG_FLAGS = [...HDR2PNG_DEFAULT_FLAGS, ...HDR2PNG_CUSTOM_FLAGS];

		// CMFT
		createProcess(CMFT_TOOL_PATH, CMFT_FLAGS)
			.then(() => {
				// HDR2PNG
				createProcess(HDR2PNG_TOOL_PATH, HDR2PNG_FLAGS)
					.then(() => {
						resolve();
					})
					.catch((error) => {
						reject(error);
					});
			});
	});
}

function generateSphericalHarmonics() {
	return new Promise((resolve, reject) => {
		const CMFT_CUSTOM_FLAGS = [
			'--input', `${input}`,
			'--filter', 'shcoeffs',
			'--outputNum', '1',
			'--output0', `${output}/sh`,
		];

		const CMFT_FLAGS = [...CMFT_DEFAULT_FLAGS, ...CMFT_CUSTOM_FLAGS];

		createProcess(CMFT_TOOL_PATH, CMFT_FLAGS)
			.then(() => {
				convertSHToBin()
					.then(() => {
						resolve();
					})
					.catch((error) => {
						reject(error);
					});
			});
	});
}

function generateMontage() {
	const MONTAGE_CUSTOM_FLAGS = [
		`${output}/level[0-7].png`,
		'-tile', '1x',
		'-geometry', '256x256',
		'-background', 'none',
		`${output}/env.png`,
	];

	const MONTAGE_FLAGS = [...MONTAGE_DEFAULT_FLAGS, ...MONTAGE_CUSTOM_FLAGS];

	createProcess(MONTAGE_TOOL_PATH, MONTAGE_FLAGS);
}

// https://m.makemepulse.com/behind-the-scenes-of-for-honor-scars-cbb800c6e05f

// Pre-calculated blinn-phong BRDF at 8 different brightness levels
// For the given gloss level of a pixel, the shader samples the corresponding version
// of the environment texture to get the amount of reflected light.

// angle = 86.77632098928787;
// getGloss(angle) ~ 1.9999999999999993

Promise.all([
	generateHighQualityLevel(),
	generateBaseLevel(), // Roughness 0
	generateGlossLevel(1, 11.038118968108332), // Roughness 0.125
	generateGlossLevel(2, 8.815820067985793), // Roughness 0.25
	generateGlossLevel(3, 7.36788005985188), // Roughness 0.375
	generateGlossLevel(4, 6.194328423796595), // Roughness 0.5
	generateGlossLevel(5, 5.103120139257955), // Roughness 0.625
	generateGlossLevel(6, 3.9191797055252344), // Roughness 0.75
	generateGlossLevel(7, 1.9662123463476706), // Roughness 0.875
	generateSphericalHarmonics(),
]).then(() => {
	generateMontage();
});
